/*
 * Mirai. Copyright (C) 2024 Vyn
 * This file is licensed under version 3 of the GNU General Public License (GPL-3.0-only)
 * The license can be found in the LICENSE file or at https://www.gnu.org/licenses/gpl-3.0.txt
 */

#pragma once

#include "date.h"
#include <optional>
#include <string>
#include <vector>

namespace mirai
{

enum task_state { TODO, DONE };

struct task_data {
	int id;
	std::string title;
	task_state state;
	std::optional<class date> due_date;
};

struct updatable_task_data {
	std::optional<std::string> title = std::nullopt;
	std::optional<task_state> state = std::nullopt;
	std::optional<std::optional<class date>> due_date = std::nullopt;
};

struct event_data {
	int id;
	class date date;
	std::string title;
	time starts_at;
	time ends_at;
};

struct updatable_event_data {
	std::optional<class date> date;
	std::optional<std::string> title;
	std::optional<time> starts_at;
	std::optional<time> ends_at;
};

class data_provider
{

  public:
	data_provider() {};
	data_provider(data_provider &) = delete;
	data_provider(data_provider &&) = delete;

	virtual ~data_provider() = default;
	virtual void save() = 0;
	virtual void load() = 0;

	virtual std::string name() const = 0;

	// Tasks
	virtual task_data insert_task(const task_data &task) = 0;
	virtual void remove_task_by_id(int taskId) = 0;
	virtual std::optional<task_data> get_task_by_id(int taskId) = 0;
	virtual std::vector<task_data> get_tasks_by_date(class date date) = 0;
	virtual std::vector<task_data> get_tasks_without_date() = 0;
	virtual std::vector<task_data> get_tasks() = 0;
	virtual void update_task(int taskId, updatable_task_data updateData) = 0;

	// Events
	virtual event_data insertEvent(const event_data &task) = 0;
	virtual void remove_event_by_id(int eventId) = 0;
	virtual std::optional<event_data> get_event_by_id(int eventId) = 0;
	virtual std::vector<event_data> get_events() = 0;
	virtual std::vector<event_data> get_events_by_date(class date date) = 0;
	virtual void update_event(int eventId, updatable_event_data updateData) = 0;

	const int id = data_provider::generate_next_id();

	void set_dirty(bool dirty)
	{
		_is_dirty = dirty;
	};

	bool is_dirty() const
	{
		return _is_dirty;
	};

  private:
	bool _is_dirty = false;

	static int generate_next_id()
	{
		static int nextId = 0;
		return nextId++;
	}
};

} // namespace mirai
