/*
 * Mirai. Copyright (C) 2024 Vyn
 * This file is licensed under version 3 of the GNU General Public License (GPL-3.0-only)
 * The license can be found in the LICENSE file or at https://www.gnu.org/licenses/gpl-3.0.txt
 */

#pragma once

#include "data_provider.h"
#include "date.h"
#include "event.h"
#include "task.h"
#include <string>
#include <vector>

namespace mirai
{

struct create_task_params {
	std::string title;
	std::optional<class date> due_date;
};

struct create_event_params {
	std::string title;
	class date date;
	time starts_at;
	time ends_at;
};

struct source_constructor {
	std::string name;
	std::string color;
	data_provider *source_data_provider;
};

class source
{
  public:
	source(source_constructor params)
		: data(params.source_data_provider), _name(params.name), _color(params.color),
		  _type("FileSystemMarkdown")
	{
	}

	source(source &) = delete;
	source(source &&) = delete;
	source operator=(source &) = delete;
	source operator=(source &&) = delete;

	~source() = default;

	void init();
	void save();
	void load();

	std::string name() const;
	std::string type() const;
	std::string color() const;
	class data_provider *data_provider();

	void set_name(const std::string &name);
	void set_color(const std::string &color);

	void create_task(const create_task_params &task);
	void remove_task(const task &task);
	std::optional<task> get_task_by_id(int taskId);
	std::vector<task> get_tasks_by_date(class date date);
	std::vector<task> get_unscheduled_tasks();
	std::vector<task> get_tasks();

	void create_event(const create_event_params &eventToCreate);
	void remove_event(const event &event);
	std::vector<class event> get_events();
	std::optional<class event> get_event_by_id(int eventId);
	std::vector<class event> get_events_by_date(class date date);

	const int id = source::generate_next_id();

	bool is_dirty() const
	{
		return data->is_dirty();
	};

  private:
	static int generate_next_id()
	{
		static int nextId = 0;
		return nextId++;
	}

	std::string _name;
	std::string _type;
	std::string _color;
	class data_provider *data;
};
} // namespace mirai
