/*
 * Mirai. Copyright (C) 2024 Vyn
 * This file is licensed under version 3 of the GNU General Public License (GPL-3.0-only)
 * The license can be found in the LICENSE file or at https://www.gnu.org/licenses/gpl-3.0.txt
 */

#include "core.h"
#include "data_provider.h"
#include "markdown_data_provider.h"
#include "rei-json/Array.h"
#include "rei-json/Object.h"
#include "rei-json/json.h"
#include "source.h"
#include <algorithm>
#include <cassert>
#include <iostream>
#include <memory>
#include <print>
#include <string>
#include <vector>

namespace mirai
{

void core::load_config(const std::string &path)
{
	std::ifstream file(path);
	if (!file) {
		return;
	}
	std::stringstream buffer;
	buffer << file.rdbuf();
	auto jsonStr = buffer.str();

	auto configJson = rei::json::parse(jsonStr);

	assert(configJson.isObject());
	assert(configJson.asObject()["files"].isArray());
	auto jsonSources = configJson.asObject()["files"].asArray();
	for (int i = 0; i < jsonSources.count(); ++i) {
		auto &filePath = jsonSources[i];
		assert(filePath.isObject());
		const auto name = filePath.asObject().getString("name");
		const auto color = filePath.asObject().getString("color");
		const auto path = filePath.asObject().getString("path");
		std::unique_ptr<data_provider> file = std::make_unique<markdown_data_provider>(path);
		data_provider *sourceDataProvider = file.release();
		sourceDataProvider->load();
		_sources.push_back(
			std::make_unique<source>(source_constructor{
				.name = name, .color = color, .source_data_provider = sourceDataProvider
			})
		);
	}
}

void core::save_config()
{
	std::ofstream file(_config_path);
	if (!file.is_open()) {
		std::print(std::cerr, "Can't save config to {}", _config_path);
		return;
	}
	auto files = rei::json::JsonArray{};
	auto configJson = rei::json::JsonObject{};
	configJson.set("files", files);
	for (auto &source : _sources) {

		rei::json::JsonObject jsonSource;
		jsonSource.set("name", source->name());
		jsonSource.set("color", source->color());
		auto dataProvider = dynamic_cast<markdown_data_provider *>(source->data_provider());
		jsonSource.set("path", dataProvider->path());
		jsonSource.set("type", "FileSystemMarkdown");
		configJson.getArray("files").push(jsonSource);
	}
	file << rei::json::toString(configJson);
	file.close();
}

core::core(const std::string &configFilePath) : _config_path(configFilePath)
{
	load_config(configFilePath);
}

void core::add_source(
	const std::string &name, const std::string &color, const std::string &type,
	std::unique_ptr<data_provider> &&new_source
)
{
	data_provider *sourceDataProvider = new_source.release();
	sourceDataProvider->load();
	_sources.push_back(
		std::make_unique<source>(source_constructor{
			.name = name, .color = color, .source_data_provider = sourceDataProvider
		})
	);
	save_config();
	source_added.emit(nullptr);
};

void core::edit_source(
	int id, const std::string &name, const std::string &color, const std::string &path
)
{
	auto source = get_source_by_id(id);
	source->set_name(name);
	source->set_color(color);

	data_provider *sourceDataProvider = source->data_provider();
	save_config();
	source_edited.emit(nullptr);
}

void core::remove_source(int id)
{
	_sources.erase(
		std::remove_if(
			_sources.begin(), _sources.end(),
			[&](const std::unique_ptr<source> &source) { return source->id == id; }
		),
		_sources.end()
	);

	save_config();
	source_removed.emit(id);
}

void core::unload_all_sources()
{
	_sources.clear();
}

void core::save()
{
	for (auto &source : _sources) {
		if (source->is_dirty()) {
			source->save();
		}
	}
}

std::vector<std::unique_ptr<source>> &core::get_sources()
{
	return _sources;
}

source *core::get_source_by_id(int id)
{
	auto source_found = std::ranges::find_if(_sources, [&](const std::unique_ptr<source> &source) {
		return source->id == id;
	});
	// assert(source_found != sources_.end()); // This should not happen
	if (source_found == _sources.end()) {
		return nullptr;
	}
	return source_found->get();
}

source *core::get_source_by_name(const std::string &name)
{
	auto source_found = std::ranges::find_if(_sources, [&](const std::unique_ptr<source> &source) {
		return source->name() == name;
	});
	assert(source_found != _sources.end()); // This should not happen
	if (source_found == _sources.end()) {
		return nullptr;
	}
	return source_found->get();
}

void core::on_source_added(std::function<void(source *)> f)
{
	source_added.register_callback(f);
}

void core::on_source_edited(std::function<void(source *)> f)
{
	source_edited.register_callback(f);
}

void core::on_source_removed(std::function<void(int)> f)
{
	source_removed.register_callback(f);
}

} // namespace mirai
