/*
 * Mirai. Copyright (C) 2024 Vyn
 * This file is licensed under version 3 of the GNU General Public License (GPL-3.0-only)
 * The license can be found in the LICENSE file or at https://www.gnu.org/licenses/gpl-3.0.txt
 */

#include "markdown_data_provider.h"
#include "cpp-utils/debug.h"
#include "data_provider.h"
#include "utils.h"
#include <algorithm>
#include <cassert>
#include <functional>
#include <optional>
#include <regex>
#include <vector>

namespace mirai
{

std::string markdown_data_provider::path() const
{
	return _file_path;
}

std::string markdown_data_provider::name() const
{
	return _data.name;
}

void markdown_data_provider::save()
{
	const std::string content = to_markdown();

	std::ofstream file(_file_path);
	if (!file.is_open()) {
		throw std::runtime_error("can't create " + _file_path);
	}

	file << content;
	file.close();
	set_dirty(false);
};

void markdown_data_provider::load()
{
	std::ifstream file(_file_path);
	if (!file.is_open()) {
		return;
	}
	std::string content = "";
	std::string line;

	while (std::getline(file, line)) {
		content += line + "\n";
	}
	file.close();
	_data = parse_markdown(content);
};

task_data markdown_data_provider::insert_task(const task_data &taskData)
{
	_data.tasks.push_back(taskData);
	set_dirty(true);
	return taskData;
}

void markdown_data_provider::update_task(int taskId, updatable_task_data updateData)
{
	auto taskData =
		ptrFindFirst(_data.tasks, [&](const task_data &task) { return task.id == taskId; });
	assert(taskData != nullptr); // Shouldn't be possible
	if (updateData.title.has_value()) {
		taskData->title = updateData.title.value();
	}
	if (updateData.state.has_value()) {
		taskData->state = updateData.state.value();
	}
	if (updateData.due_date.has_value()) {
		taskData->due_date = updateData.due_date.value();
	}
	set_dirty(true);
}

std::optional<task_data> markdown_data_provider::get_task_by_id(int taskId)
{
	return findFirst(_data.tasks, [&](const task_data &task) { return task.id == taskId; });
}

std::vector<task_data> markdown_data_provider::get_tasks_without_date()
{
	return findAll(_data.tasks, [](const task_data &t) { return t.due_date == std::nullopt; });
}

std::vector<task_data> markdown_data_provider::get_tasks_by_date(class date date)
{
	return findAll(_data.tasks, [&](const task_data &t) { return t.due_date == date; });
}

std::vector<task_data> markdown_data_provider::get_tasks()
{
	return _data.tasks;
}

void markdown_data_provider::remove_task_by_id(int taskId)
{
	_data.tasks.erase(
		std::remove_if(
			_data.tasks.begin(), _data.tasks.end(),
			[&](const task_data &task) { return task.id == taskId; }
		),
		_data.tasks.end()
	);
	set_dirty(true);
}

event_data markdown_data_provider::insertEvent(const event_data &eventData)
{
	_data.events.push_back(eventData);
	set_dirty(true);
	return eventData;
}

void markdown_data_provider::update_event(int eventId, updatable_event_data updateData)
{
	auto eventData =
		ptrFindFirst(_data.events, [&](const event_data &event) { return event.id == eventId; });
	assert(eventData != nullptr); // Shouldn't be possible
	if (updateData.title.has_value()) {
		eventData->title = updateData.title.value();
	}
	if (updateData.date.has_value()) {
		eventData->date = updateData.date.value();
	}
	if (updateData.starts_at.has_value()) {
		eventData->starts_at = updateData.starts_at.value();
	}
	if (updateData.ends_at.has_value()) {
		eventData->ends_at = updateData.ends_at.value();
	}
	set_dirty(true);
}

void markdown_data_provider::remove_event_by_id(int eventId)
{
	_data.events.erase(
		std::remove_if(
			_data.events.begin(), _data.events.end(),
			[&](const event_data &event) { return event.id == eventId; }
		),
		_data.events.end()
	);
	set_dirty(true);
}

std::optional<event_data> markdown_data_provider::get_event_by_id(int eventId)
{
	return findFirst(_data.events, [&](const event_data &event) { return event.id == eventId; });
}

std::vector<event_data> markdown_data_provider::get_events_by_date(class date date)
{
	return findAll(_data.events, [&](const event_data &event) { return event.date == date; });
}

std::vector<event_data> markdown_data_provider::get_events()
{
	return _data.events;
}
} // namespace mirai
