/*
 * Mirai. Copyright (C) 2024 Vyn
 * This file is licensed under version 3 of the GNU General Public License (GPL-3.0-only)
 * The license can be found in the LICENSE file or at https://www.gnu.org/licenses/gpl-3.0.txt
 */

#include "markdown_data_provider.h"
#include "cpp-utils/debug.h"
#include "data_provider.h"
#include "utils.h"
#include <optional>
#include <print>
#include <regex>

namespace mirai
{

/*Tags TodoMdFormat::extractTagsFromMetadata(std::string metadata)*/
/*{*/
/*Tags tags;*/

/*std::regex regex("(#([a-zA-Z0-1]+))");*/
/*std::smatch matches;*/

/*while (std::regex_search(metadata, matches, regex)) {*/
/*if (!vectorUtils::contains(tags, matches[2].str())) {*/
/*tags.push_back(matches[2]);*/
/*}*/
/*metadata = matches.suffix();*/
/*}*/
/*return tags;*/
/*}*/

std::string fieldWithSpace(const std::string &field)
{
	if (field.length() == 0) {
		return "";
	}
	return (field + " ");
}

task_data stringToTask(const std::string &str, const std::string &date)
{
	std::smatch matches;
	std::regex regex(
		"- \\[(\\s|X)\\] (([0-9]{2}:[0-9]{2})-([0-9]{2}:[0-9]{2}) > )?(.*?)( -- (.*))?"
	);
	std::regex_match(str, matches, regex);

	std::string text = stringUtils::trim(matches[5]);

	task_data taskItem{
		.title = text,
		.state = str.substr(0, 5) == "- [X]" ? DONE : TODO,
	};
	return taskItem;
}

time stringToTime(const std::string &str)
{
	if (str.length() != 5) {
		throw std::runtime_error("Str time length must be 5 (got '" + str + "')");
	}
	if (str[2] != 'h') {
		throw std::runtime_error("Malformated time range");
	}
	int hour, minute;
	std::from_chars(str.data(), str.data() + 2, hour);
	std::from_chars(str.data() + 3, str.data() + 5, minute);
	time time{hour, minute};

	return time;
}

event_data stringToEvent(const std::string &str, const std::string &dateString)
{
	std::smatch matches;
	std::regex regex("> (([0-9]{2}h[0-9]{2})-([0-9]{2}h[0-9]{2}) )(.*?)( -- (.*))?");
	std::regex_match(str, matches, regex);

	std::string text = stringUtils::trim(matches[4]);

	auto date = string_to_date(dateString);

	if (!date.has_value()) {
		throw std::runtime_error("Malformated date");
	}

	event_data eventData{
		.date = date.value(),
		.title = text,
		.starts_at = stringToTime(matches[2]),
		.ends_at = stringToTime(matches[3]),
	};
	return eventData;
}

std::string taskToString(const task_data &task)
{
	std::string str = task.title;
	str = (task.state == DONE ? "- [X] " : "- [ ] ") + str;
	return str;
}

std::string markdown_data_provider::to_markdown()
{
	std::string result = "# " + name() + "\n\n";
	std::string currentDate = "";

	int next_task_index = 0;
	int next_event_index = 0;

	std::sort(_data.tasks.begin(), _data.tasks.end(), [](const task_data &t1, const task_data &t2) {
		if (!t1.due_date.has_value() && !t2.due_date.has_value()) {
			return false;
		}
		if (t1.due_date.has_value() && !t2.due_date.has_value()) {
			return true;
		}
		if (!t1.due_date.has_value() && t2.due_date.has_value()) {
			return false;
		}
		return t1.due_date < t2.due_date;
	});

	while (next_event_index < _data.events.size() || next_task_index < _data.tasks.size()) {
		std::optional<class date> current_date = std::nullopt;
		// First we find the lowest date between the events and the tasks
		if (_data.events.size() > 0 && next_event_index < _data.events.size()) {
			current_date = _data.events.at(next_event_index).date;
		}
		if (_data.tasks.size() > 0 && next_task_index < _data.tasks.size()) {
			auto &task = _data.tasks.at(next_task_index);
			if ((task.due_date.has_value() && current_date.has_value()
				 && task.due_date < current_date)
				|| !current_date.has_value()) {
				current_date = _data.tasks.at(next_task_index).due_date;
			}
		}

		if (current_date.has_value()) {
			result += "## "
					+ std::format(
						  "{:02d}-{:02d}-{:02d}", current_date.value().year,
						  current_date.value().month, current_date.value().day
					)
					+ "\n\n";
		} else {
			result += "## Unscheduled\n\n";
		}
		bool should_add_newline_before_tasks = false;
		while (next_event_index < _data.events.size()) {
			auto &current_event = _data.events.at(next_event_index);
			if (current_event.date != current_date) {
				break;
			}
			should_add_newline_before_tasks = true;
			auto &start = current_event.starts_at;
			auto &end = current_event.ends_at;
			result += "> "
					+ std::format(
						  "{:02d}h{:02d}-{:02d}h{:02d} {}", start.hour, start.minute, end.hour,
						  end.minute, current_event.title
					)
					+ "\n";
			next_event_index++;
		}

		if (should_add_newline_before_tasks) {
			result += "\n";
		}

		while (next_task_index < _data.tasks.size()) {
			auto &current_task = _data.tasks.at(next_task_index);
			if (current_task.due_date != current_date) {
				break;
			}
			result += taskToString(current_task) + "\n";
			next_task_index++;
		}
		result += '\n';
	}

	return result;
};

markdown_data markdown_data_provider::parse_markdown(const std::string &content)
{
	cpputils::debug::Timer readMdFormatDuration;

	std::string line;

	std::stringstream contentStream(content);

	if (!std::getline(contentStream, line) || line.substr(0, 2) != "# ") {
		std::println("Couldn't find the task list name");
	}

	_data.name = line.substr(2);
	std::string currentDateString = "";

	std::optional<class date> current_date = std::nullopt;

	cpputils::debug::Timer stringToTaskDuration;
	stringToTaskDuration.reset();
	cpputils::debug::Timer gelinesDuration;
	while (std::getline(contentStream, line)) {
		if (std::string_view{line.data(), 3} == "## ") {
			currentDateString = line.substr(3);
			if (currentDateString == "Unscheduled") {
				current_date = std::nullopt;
				continue;
			}
			current_date = mirai::string_to_date(currentDateString);
			if (!current_date.has_value()) {
				throw std::runtime_error("Malformated date (1)");
			}
		} else if (line.starts_with("> ")) {
			auto event = stringToEvent(line, currentDateString);
			if (current_date.has_value()) {
				insertEvent(
					{.id = generate_unique_id(),
					 .date = current_date.value(),
					 .title = event.title,
					 .starts_at = event.starts_at,
					 .ends_at = event.ends_at}
				);
			}
		} else if (line.starts_with("- [ ]") || line.starts_with("- [X]")) {
			stringToTaskDuration.start();
			task_data taskItemData = stringToTask(line, currentDateString);
			stringToTaskDuration.stop();
			insert_task(
				{.id = generate_unique_id(),
				 .title = taskItemData.title,
				 .state = taskItemData.state,
				 .due_date = current_date}
			);
		}
	}

	gelinesDuration.printTimeElapsed("getlinesDuration");
	stringToTaskDuration.printTimeElapsed("stringToTaskDuration");
	readMdFormatDuration.printTimeElapsed("Reading MD File duration");
	return _data;
}
} // namespace mirai
