/*
 * Mirai. Copyright (C) 2024 Vyn
 * This file is licensed under version 3 of the GNU General Public License (GPL-3.0-only)
 * The license can be found in the LICENSE file or at https://www.gnu.org/licenses/gpl-3.0.txt
 */

#include "source.h"
#include "data_provider.h"
#include "utils.h"
#include <algorithm>
#include <iterator>
#include <optional>
#include <string>
#include <vector>

namespace mirai
{

void source::save()
{
	data->save();
}

void source::create_task(const create_task_params &task)
{
	data->insert_task({
		.id = generate_unique_id(),
		.title = task.title,
		.state = TODO,
		.due_date = task.due_date,
	});
};

void source::create_event(const create_event_params &new_event)
{
	data->insertEvent({
		.id = generate_unique_id(),
		.date = new_event.date,
		.title = new_event.title,
		.starts_at = new_event.starts_at,
		.ends_at = new_event.ends_at,
	});
};

std::vector<class event> source::get_events()
{
	auto event_data = data->get_events();
	std::vector<class event> events;
	std::transform(
		event_data.begin(), event_data.end(), std::back_inserter(events),
		[&](const struct event_data &event_data) { return event{data, event_data}; }
	);
	return events;
}

std::optional<class event> source::get_event_by_id(int evendId)
{
	auto event_data = data->get_event_by_id(evendId);
	if (!event_data.has_value()) {
		return std::nullopt;
	}
	return event{data, event_data.value()};
}

void source::remove_event(const event &event)
{
	data->remove_event_by_id(event.id());
}

void source::remove_task(const task &task)
{
	data->remove_task_by_id(task.id());
}

std::optional<task> source::get_task_by_id(int taskId)
{
	auto taskData = data->get_task_by_id(taskId);
	if (!taskData.has_value()) {
		return std::nullopt;
	}
	task task{data, taskData.value()};
	return task;
}

std::vector<task> source::get_tasks()
{
	auto tasksData = data->get_tasks();
	std::vector<task> tasks;
	std::transform(
		tasksData.begin(), tasksData.end(), std::back_inserter(tasks),
		[&](const task_data &taskData) { return task{data, taskData}; }
	);
	return tasks;
}

std::vector<task> source::get_unscheduled_tasks()
{
	auto tasksData = data->get_tasks_without_date();
	std::vector<task> tasks;
	std::transform(
		tasksData.begin(), tasksData.end(), std::back_inserter(tasks),
		[&](const task_data &taskData) { return task{data, taskData}; }
	);
	return tasks;
}

std::string source::name() const
{
	return _name;
}

std::string source::color() const
{
	return _color;
}

void source::set_name(const std::string &name)
{
	_name = name;
}

void source::set_color(const std::string &color)
{
	_color = color;
}

std::string source::type() const
{
	// There is only 1 type for now
	return _type;
}

data_provider *source::data_provider()
{
	return data;
}
} // namespace mirai
